local Utopia = LibStub("AceAddon-3.0"):GetAddon("Utopia", true)
if (not Utopia) then
	return
end
Utopia:UpdateVersion("$Revision: 47 $")

local isWoW3dot1 = select(2,GetBuildInfo()) >= "9614"

local L = LibStub("AceLocale-3.0"):GetLocale("Utopia")

-- All talent scans are done by name. No messing around with slot indexes etc.

local bloodlustID = UnitFactionGroup("player") == "Horde" and 2825 or 32182

local statMult, wolfHowl, armorMajor, armorMinor, unleashedRage
if (isWoW3dot1) then
	statMult = {
		icon = select(3, GetSpellInfo(20217)),
		attribute = L["Stats"],
		mode = "Stats",
		spells = {
			[GetSpellInfo(20217)] = {						-- Blessing of Kings
				alternate = GetSpellInfo(25898),			-- Greater Blessing of Kings
				class = "PALADIN",
				amount = 10,
				amountType = "%",
			},
		},
	}
	unleashedRage = {
		class = "SHAMAN",
		spec = "Enhancement",
		amounts = {4, 7, 10},
		maxTalentPoints = 3,
		amountType = "%",
		requiredTalent = GetSpellInfo(30809),
	}
	armorMajor = {
		icon = select(3, GetSpellInfo(47467)),
		attribute = L["Armor"],
		spells = {
			[GetSpellInfo(8647)] = {						-- Expose Armor
				class = "ROGUE",
				amount = 20,
				amountType = "%",
				--rankLevels = {14, 26, 36, 46, 56, 66, 77},
				--amountsPerStack = {90, 180, 270, 360, 450, 520, 785},
			},
			[GetSpellInfo(47467)] = {						-- Sunder Armor
				class = "WARRIOR",
				amountPerStack = 4,
				amountType = "%",
				maxStacks = 5,
				--rankLevels = {10, 22, 34, 46, 58, 67, 77},
				--amountsPerStack = {90, 180, 270, 360, 450, 520, 785},
				--maxStacks = 5
			},
			--[GetSpellInfo(55754)] = {						-- Acid Spit
			--	class = "HUNTER",
			--	rankLevels = {1, 16, 32, 48, 64, 80},
			--	amountsPerStack = {10, 120, 250, 360, 520, 785},
			--	maxStacks = 5,
			--	pet = L["Worm"]
			--},
		},
	}
	armorMinor = {
		icon = select(3, GetSpellInfo(770)),
		attribute = L["Armor"],
		spells = {
			[GetSpellInfo(16231)] = {						-- Curse of Recklessness
				class = "WARLOCK",
				amount = 5,
				amountType = "%",
				--rankLevels = {14, 28, 42, 56, 69, 79},
				--amounts = {175, 285, 395, 505, 610, 1260},
			},
			[GetSpellInfo(770)] = {						-- Faerie Fire
				class = "DRUID",
				amount = 5,
				amountType = "%",
				--rankLevels = {18, 30, 42, 54, 66, 76},
				--amounts = {175, 285, 395, 505, 610, 1260},
			},
			[GetSpellInfo(16857)] = {						-- Faerie Fire (Feral)
				class = "DRUID",
				amount = 5,
				amountType = "%",
				--rankLevels = {18, 30, 42, 54, 66, 76},
				--amounts = {175, 285, 395, 505, 610, 1260},
			},
			[GetSpellInfo(56631)] = {						-- Sting
				class = "HUNTER",
				amount = 5,
				amountType = "%",
				--rankLevels = {1, 16, 32, 48, 64, 80},
				--amounts = {175, 285, 395, 505, 610, 1260},
				pet = L["Wasp"]
			},
		},
	}
else
	statMult = {
		icon = select(3, GetSpellInfo(20217)),
		attribute = L["Stats"],
		mode = "Stats",
		spells = {
			[GetSpellInfo(20217)] = {						-- Blessing of Kings
				alternate = GetSpellInfo(25898),			-- Greater Blessing of Kings
				class = "PALADIN",
				spec = "Protection",
				amount = 2,
				amountType = "%",
				improvedPercentagePerTalentPoint = 2,
				maxTalentPoints = 4,
				requiredTalent = GetSpellInfo(20217),		-- Blessing of Kings
				improved = GetSpellInfo(59295),				-- Improved Blessing of Kings
			},
		},
	}
	wolfHowl = {						-- Furious Howl
		class = "HUNTER",
		pet = L["Wolf"],
		amount = 2,				-- AP value = 2 x [Pet Level] ?
		amountFunc = function(src) return (UnitLevel(src or "player") - 1) * 2 end,
	}
	unleashedRage = {
		class = "SHAMAN",
		spec = "Enhancement",
		amountPerTalentPoint = 2,
		maxTalentPoints = 5,
		amountType = "%",
		requiredTalent = GetSpellInfo(30809),
	}
	armorMajor = {
		icon = select(3, GetSpellInfo(47467)),
		attribute = L["Armor"],
		spells = {
			[GetSpellInfo(48669)] = {						-- Expose Armor
				class = "ROGUE",
				rankLevels = {14, 26, 36, 46, 56, 66, 77},
				amountsPerStack = {90, 180, 270, 360, 450, 520, 785},
				maxStacks = 5,
				assumeMaxStacks = true,
				stacksDescription = L["combos"],
				-- NOTE: Expose Armor doesn't stack, but rather has the same debuff used
				-- whether they use 1 or 5 Combos. There is no way to determine this currently.
			},
			[GetSpellInfo(47467)] = {						-- Sunder Armor
				class = "WARRIOR",
				rankLevels = {10, 22, 34, 46, 58, 67, 77},
				amountsPerStack = {90, 180, 270, 360, 450, 520, 785},
				maxStacks = 5
			},
			[GetSpellInfo(55754)] = {						-- Acid Spit
				class = "HUNTER",
				rankLevels = {1, 16, 32, 48, 64, 80},
				amountsPerStack = {10, 120, 250, 360, 520, 785},
				maxStacks = 5,
				pet = L["Worm"]
			},
		},
	}
	armorMinor = {
		icon = select(3, GetSpellInfo(48476)),
		attribute = L["Armor"],
		spells = {
			[GetSpellInfo(57595)] = {						-- Curse of Recklessness
				class = "WARLOCK",
				rankLevels = {14, 28, 42, 56, 69, 79},
				amounts = {175, 285, 395, 505, 610, 1260},
			},
			[GetSpellInfo(48476)] = {						-- Faerie Fire
				class = "DRUID",
				rankLevels = {18, 30, 42, 54, 66, 76},
				amounts = {175, 285, 395, 505, 610, 1260},
			},
			[GetSpellInfo(48475)] = {						-- Faerie Fire (Feral)
				class = "DRUID",
				rankLevels = {18, 30, 42, 54, 66, 76},
				amounts = {175, 285, 395, 505, 610, 1260},
			},
			[GetSpellInfo(56631)] = {						-- Sting
				class = "HUNTER",
				rankLevels = {1, 16, 32, 48, 64, 80},
				amounts = {175, 285, 395, 505, 610, 1260},
				pet = L["Wasp"]
			},
		},
	}
end

Utopia.buffs = {
	[L["Armor (%)"]] = {
		icon = select(3, GetSpellInfo(16240)),
		attribute = L["Armor"],
		mode = "Mitigation",
		spells = {
			[GetSpellInfo(16240)] = {						-- Ancestral Healing
				class = "SHAMAN",
				amountPerTalentPoint = 25/3,
				maxTalentPoints = 3,
				amountType = "%",
				requiredTalent = GetSpellInfo(16240),		-- Ancestral Healing
			},
			[GetSpellInfo(15363)] = {						-- Inspiration
				class = "PRIEST",
				amountPerTalentPoint = 25/3,
				maxTalentPoints = 3,
				amountType = "%",
				requiredTalent = GetSpellInfo(15363),		-- Inspiration
			},
		},
	},
	[L["Attack Power"]] = {
		icon = select(3, GetSpellInfo(47436)),
		attribute = L["AP"],
		mode = "DPS",
		spells = {
			[GetSpellInfo(47436)] = {						-- Battle Shout
				class = "WARRIOR",
				rankLevels = {1, 12, 22, 32, 42, 52, 60, 69, 78},
				amounts = {15, 35, 55, 85, 130, 185, 232, 305, 548},
				improvedPercentagePerTalentPoint = 5,
				maxTalentPoints = 5,
				improved = GetSpellInfo(12861),				-- Commanding Presence
			},
			[GetSpellInfo(48932)] = {						-- Blessing of Might
				alternate = GetSpellInfo(27141),			-- Greater Blessing of Might
				alternateOffset = 5,						-- How many ranks to add if it's the alternate spell
				class = "PALADIN",
				rankLevels = {4, 12, 22, 32, 42, 52, 60, 70, 73, 79},
				amounts = {20, 35, 55, 85, 130, 185, 232, 306, 410, 550},
				improvedPercentagePerTalentPoint = 25/2,
				maxTalentPoints = 2,
				improved = GetSpellInfo(20045),				-- Improved Blessing of Might
			},
			[GetSpellInfo(24604)] = wolfHowl,
		},
	},
	[L["Attack Power (%)"]] = {
		icon = select(3, GetSpellInfo(19506)),
		attribute = L["AP"],
		mode = "DPS",
		spells = {
			[GetSpellInfo(19506)] = {						-- Trueshot Aura
				class = "HUNTER",
				spec = "Marksmanship",
				amount = 10,
				amountType = "%",
				requiredTalent = GetSpellInfo(19506),
			},
			[GetSpellInfo(53138)] = {						-- Abomination's Might
				class = "DEATHKNIGHT",
				spec = "Blood",
				amount = 10,
				amountType = "%",
				requiredTalent = GetSpellInfo(53138),
			},
			[GetSpellInfo(30809)] = unleashedRage,
		},
	},
	["Bloodlust/Heroism"] = {
		icon = select(3, GetSpellInfo(bloodlustID)),
		attribute = L["Haste"],
		mode = "DPS",
		spells = {
			[GetSpellInfo(bloodlustID)] = {					-- Bloodlust / Heroism
				class = "SHAMAN",
				amount = 30,
				amountType = "%",
			},
		},
	},
	[L["Crit Chance (Melee)"]] = {
		icon = select(3, GetSpellInfo(17007)),
		attribute = L["Melee Crit"],
		mode = "Stats",
		spells = {
			[GetSpellInfo(17007)] = {						-- Leader of the Pack
				class = "DRUID",
				spec = "Feral",
				amount = 5,
				amountType = "%",
				requiredTalent = GetSpellInfo(17007),		-- Leader of the Pack
			},
			[GetSpellInfo(29801)] = {						-- Rampage
				class = "WARRIOR",
				spec = "Fury",
				amount = 5,
				amountType = "%",
				requiredTalent = GetSpellInfo(29801),		-- Rampage
			},
		},
	},
	[L["Crit Chance (Spell)"]] = {
		icon = select(3, GetSpellInfo(24907)),
		attribute = L["Spell Crit"],
		mode = "Stats",
		spells = {
			[GetSpellInfo(24907)] = {						-- Moonkin Aura
				class = "DRUID",
				spec = "Balance",
				amount = 5,
				amountType = "%",
				requiredTalent = GetSpellInfo(24858),		-- Moonkin Form
			},
			[GetSpellInfo(51470)] = {						-- Elemental Oath
				class = "SHAMAN",
				spec = "Elemental",
				amountPerTalentPoint = 2.5,
				maxTalentPoints = 2,
				amountType = "%",
				requiredTalent = GetSpellInfo(51470),		-- Elemental Oath
			},
		},
	},
	[L["Damage (%)"]] = {
		icon = select(3, GetSpellInfo(34460)),
		attribute = L["Damage Modifier"],
		mode = "DPS",
		spells = {
			[GetSpellInfo(34460)] = {						-- Ferocious Inspiration
				class = "SHAMAN",
				spec = "Enhancement",
				amountPerTalentPoint = 1,
				maxTalentPoints = 3,
				amountType = "%",
				temp = true,
				requiredTalent = GetSpellInfo(34460),
			},
			[GetSpellInfo(31869)] = {						-- Sanctified Retribution
				class = "PALADIN",
				spec = "Retribution",
				amount = 3,
				amountType = "%",
				requiredTalent = GetSpellInfo(31869),
			},
		},
	},
	[L["Damage Reduction (%)"]] = {
		icon = select(3, GetSpellInfo(20911)),
		attribute = L["Damage Reduction"],
		mode = "Mitigation",
		spells = {
			[GetSpellInfo(20911)] = {						-- Blessing of Sanctuary
				alternate = GetSpellInfo(25899),			-- Greater Blessing of Sanctuary
				class = "PALADIN",
				spec = "Protection",
				amount = 3,
				amountType = "%",
				requiredTalent = GetSpellInfo(20911),
			},
			[GetSpellInfo(47517)] = {						-- Grace
				class = "PRIEST",
				spec = "Discipline",
				amount = 1,
				amountType = "%",
				temp = true,
				requiredTalent = GetSpellInfo(47517),
			},
		},
	},
	[L["Haste"]] = {
		icon = select(3, GetSpellInfo(53648)),
		attribute = L["Haste"],
		mode = "DPS",
		spells = {
			[GetSpellInfo(53648)] = {						-- Swift Retribution
				class = "PALADIN",
				spec = "Retribution",
				amountPerTalentPoint = 1,
				maxTalentPoints = 3,
				amountType = "%",
				requiredTalent = GetSpellInfo(53648),
			},
			[GetSpellInfo(48396)] = {						-- Improved Moonkin Form
				class = "DRUID",
				spec = "Moonkin",
				amountPerTalentPoint = 1,
				maxTalentPoints = 3,
				amountType = "%",
				requiredTalent = GetSpellInfo(48396),
			},
		},
	},
	[L["Haste (Melee)"]] = {
		icon = select(3, GetSpellInfo(55610)),
		attribute = L["Haste"],
		mode = "DPS",
		spells = {
			[GetSpellInfo(55610)] = {						-- Improved Icy Talons
				class = "DEATHKNIGHT",
				spec = "Frost",
				amount = 20,
				amountType = "%",
				temp = true,
				requiredTalent = GetSpellInfo(55610),
			},
			[GetSpellInfo(8512)] = {						-- Windfury Totem
				class = "SHAMAN",
				totem = "air",
				amount = 16,
				amountType = "%",
				improvedPercentagePerTalentPoint = 2,
				maxTalentPoints = 2,
				improved = GetSpellInfo(29193),				-- Improved Windfury Totem
			},
		},
	},
	[L["Haste (Spell)"]] = {
		icon = select(3, GetSpellInfo(3738)),
		attribute = L["Haste"],
		mode = "DPS",
		spells = {
			[GetSpellInfo(3738)] = {						-- Wrath of Air Totem
				class = "SHAMAN",
				totem = "air",
				amount = 5,
				amountType = "%",
			},
		},
	},
	[L["Healing Received"]] = {
		icon = select(3, GetSpellInfo(33891)),
		attribute = L["Healing"],
		mode = "Healing",
		spells = {
			[GetSpellInfo(33891)] = {						-- Tree of Life
				class = "DRUID",
				spec = "Restoration",
				amount = 6,
				amountType = "%",
				requiredTalent = GetSpellInfo(33891),
			},
			[GetSpellInfo(20140)] = {						-- Improved Devotion Aura
				class = "PALADIN",
				spec = "Protection",
				amountPerTalentPoint = 2,
				maxTalentPoints = 3,
				amountType = "%",
				requiredTalent = GetSpellInfo(20140),
			},
		},
	},
	[L["Health"]] = {
		icon = select(3, GetSpellInfo(47440)),
		attribute = L["Health"],
		mode = "Health",
		spells = {
			[GetSpellInfo(47982)] = {						-- Blood Pact
				class = "WARLOCK",
				pet = L["Imp"],
				rankLevels = {4, 14, 26, 38, 50, 62, 74},
				amounts = {20, 70, 160, 270, 380, 660, 1330},
			},
			[GetSpellInfo(47440)] = {						-- Commanding Shout
				class = "WARRIOR",
				rankLevels = {68, 74, 80},
				amounts = {1080, 1855, 2255},
				improvedPercentagePerTalentPoint = 5,
				maxTalentPoints = 5,
				improved = GetSpellInfo(12861),				-- Commanding Presence
			},
		},
	},
	[L["Intellect"]] = {
		icon = select(3, GetSpellInfo(42995)),
		attribute = L["Intellect"],
		mode = "Stats",
		spells = {
			[GetSpellInfo(42995)] = {						-- Arcane Intellect
				alternate = GetSpellInfo(27127),			-- Arcane Brilliance
				alternateOffset = 4,						-- How many ranks to add if it's the alternate spell
				class = "MAGE",
				rankLevels = {1, 14, 28, 42, 56, 70, 80},
				amounts = {2, 7, 15, 22, 31, 40, 60},
			},
			[GetSpellInfo(61024)] = {						-- Dalaran Intellect
				alternate = GetSpellInfo(61316),			-- Dalaran Brilliance
				class = "MAGE",
				rankLevels = {80},
				amounts = {60},
			},
			[GetSpellInfo(57567)] = {						-- Fel Intelligence
				class = "WARLOCK",
				pet = "Felhunter",
				rankLevels = {32, 42, 52, 62, 72},
				amounts = {12, 18, 25, 32, 48},
				improvedPercentagePerTalentPoint = 5,
				maxTalentPoints = 2,
				improved = GetSpellInfo(54038),				-- Improved Felhunter
			},
		},
	},
	[L["Mana Regen"]] = {
		icon = select(3, GetSpellInfo(48936)),
		attribute = L["mp5"],
		mode = "mp5",
		spells = {
			[GetSpellInfo(48936)] = {						-- Blessing of Wisdom
				alternate = GetSpellInfo(27143),			-- Greater Blessing of Wisdom
				alternateOffset = 4,						-- How many ranks to add if it's the alternate spell
				class = "PALADIN",
				rankLevels = {14, 24, 34, 44, 54, 60, 65, 71, 77},
				amounts = {10, 15, 20, 25, 30, 33, 41, 73, 91},
				improvedPercentagePerTalentPoint = 10,
				maxTalentPoints = 2,
				improved = GetSpellInfo(20245),				-- Improved Blessing of Wisdom
			},
		},
	},
	[L["Mana Regen (Pet)"]] = {
		icon = select(3, GetSpellInfo(44561)),
		attribute = L["Mana"],
		mode = "mp5",
		spells = {
			[GetSpellInfo(44561)] = {						-- Improved Water Elemental
				class = "MAGE",
				spec = "Frost",
				amountPerTalentPoint = 0.2,
				maxTalentPoints = 3,
				amountType = "%",
				requiredTalent = GetSpellInfo(44561),		-- Improved Water Elemental
			},
		},
	},
	[GetSpellInfo(57669)] = {								-- Replenishment
		icon = select(3, GetSpellInfo(57669)),
		id = 57669,
		attribute = L["Mana Regen"],
		mode = "mp5",
		spells = {
			[GetSpellInfo(53292)] = {						-- Hunting Party
				alternate = GetSpellInfo(57669),
				class = "HUNTER",
				spec = "Survival",
				requiredTalent = GetSpellInfo(53292),		-- Hunting Party
			},
			[GetSpellInfo(31878)] = {						-- Judgements of the Wise
				alternate = GetSpellInfo(57669),
				class = "PALADIN",
				spec = "Retribution",
				requiredTalent = GetSpellInfo(31878),		-- Judgements of the Wise
			},
			[GetSpellInfo(48160)] = {						-- Vampiric Touch
				alternate = GetSpellInfo(57669),
				class = "PRIEST",
				spec = "Shadow",
				requiredTalent = GetSpellInfo(48160),		-- Vampiric Touch
			},
		},
	},
	[L["Spell Power"]] = {
		icon = select(3, GetSpellInfo(57722)),
		attribute = L["Spell Power"],
		mode = "DPS",
		spells = {
			[GetSpellInfo(47240)] = {						-- Demonic Pact (Temp Buff)
				class = "WARLOCK",
				pet = "any",
				amountPerTalentPoint = 2,
				maxTalentPoints = 5,
				amountType = "%",
				temp = true,
				requiredTalent = GetSpellInfo(47240),		-- Demonic Pact
			},
			[GetSpellInfo(58656)] = {						-- Flametongue Totem
				class = "SHAMAN",
				totem = "fire",
				rankLevels = {28, 38, 48, 58, 67, 71, 75, 80},
				amounts = {25, 37, 49, 62, 73, 106, 122, 144},
				improvedPercentagePerTalentPoint = 5,
				maxTalentPoints = 3,
				improved = GetSpellInfo(52456),				-- Enhancing Totems
			},
			[GetSpellInfo(57722)] = {						-- Totem of Wrath
				class = "SHAMAN",
				totem = "fire",
				rankLevels = {50, 60, 70, 80},
				amounts = {100, 120, 140, 280},
				requiredTalent = GetSpellInfo(57722),		-- Totem of Wrath
			},
			[GetSpellInfo(33182)] = {						-- Improved Divine Spirit
				class = "PRIEST",
				spec = "Discipline",
				amounts = {17, 23, 33, 40, 50, 80},			-- TODO: This currently assumes 2/2 points in Improved Divine Spirit
				requiredTalent = GetSpellInfo(33182),
			},
		},
	},
	[L["Spirit"]] = {
		icon = select(3, GetSpellInfo(48073)),
		attribute = L["Spirit"],
		mode = "Stats",
		spells = {
			[GetSpellInfo(48073)] = {						-- Divine Spirit
				alternate = GetSpellInfo(32999),			-- Prayer of Spirit
				alternateOffset = 3,						-- How many ranks to add if it's the alternate spell
				class = "PRIEST",
				rankLevels = {30, 40, 50, 60, 70, 80},
				amounts = {17, 23, 33, 40, 50, 80},
				requiredTalent = GetSpellInfo(48073),		-- Divine Spirit
			},
			[GetSpellInfo(57567)] = {						-- Fel Intelligence
				class = "WARLOCK",
				pet = "Felhunter",
				rankLevels = {32, 42, 52, 62, 72},
				amounts = {18, 26, 32, 40, 64},
				improvedPercentagePerTalentPoint = 5,
				maxTalentPoints = 2,
				improved = GetSpellInfo(54038),				-- Improved Felhunter
			},
		},
	},
	[L["Stamina"]] = {
		icon = select(3, GetSpellInfo(48161)),
		attribute = L["Stamina"],
		mode = "Stats",
		spells = {
			[GetSpellInfo(48161)] = {						-- Power Word: Fortitude
				alternate = GetSpellInfo(25392),			-- Prayer of Fortitude
				alternateOffset = 4,						-- How many ranks to add if it's the alternate spell
				class = "PRIEST",
				rankLevels = {1, 12, 24, 36, 48, 60, 70, 80},
				amounts = {3, 8, 20, 32, 43, 54, 79, 165},
				improvedPercentagePerTalentPoint = 15,
				maxTalentPoints = 2,
				improved = GetSpellInfo(14767),				-- Improved Power Word: Fortitude
			},
		},
	},
	[L["Stat Add"]] = {
		icon = select(3, GetSpellInfo(48469)),
		attribute = L["Stats"],
		mode = "Stats",
		spells = {
			[GetSpellInfo(48469)] = {						-- Mark of the Wild
				alternate = GetSpellInfo(26991),			-- Gift of the Wild
				alternateOffset = 5,						-- How many ranks to add if it's the alternate spell
				class = "DRUID",
				rankLevels = {1, 10, 20, 30, 40, 50, 60, 70, 80},
				amounts = {0, 2, 4, 6, 8, 10, 12, 14, 37},
				improvedPercentagePerTalentPoint = 20,
				maxTalentPoints = 2,
				improved = GetSpellInfo(17051),				-- Improved Mark of the Wild
			},
		},
	},
	[L["Stat Multiplier"]] = statMult,
	[L["Strength & Agility"]] = {
		icon = select(3, GetSpellInfo(58643)),
		attribute = L["Str & Agi"],
		mode = "DPS",
		spells = {
			[GetSpellInfo(58643)] = {						-- Strength of Earth Totem
				class = "SHAMAN",
				totem = "earth",
				totemBuffName = GetSpellInfo(58646),
				rankLevels = {10, 24, 38, 52, 60, 65, 75, 80},
				amounts = {10, 20, 36, 61, 77, 86, 115, 155},
				improvedPercentagePerTalentPoint = 5,
				maxTalentPoints = 3,
				improved = GetSpellInfo(52456),				-- Enhancing Totems
			},
			[GetSpellInfo(57623)] = {						-- Horn of Winter
				class = "DEATHKNIGHT",
				rankLevels = {65, 75},
				amounts = {86, 155},
			},
		},
	},
}

Utopia.debuffs = {
	[L["Armor (Major)"]] = armorMajor,
	[L["Armor (Minor)"]] = armorMinor,
	[L["Attack Power"]] = {
		icon = select(3, GetSpellInfo(47437)),
		attribute = L["Attack Power"],
		spells = {
			[GetSpellInfo(50511)] = {						-- Curse of Weakness
				class = "WARLOCK",
				rankLevels = {4, 12, 22, 32, 42, 52, 61, 69, 71},
				amounts = {21, 41, 64, 82, 123, 163, 257, 350, 478},
				spec = "Affliction",
				improvedPercentagePerTalentPoint = 10,
				maxTalentPoints = 2,
				improved = GetSpellInfo(18179)				-- Frailty
			},
			[GetSpellInfo(48560)] = {						-- Demoralizing Roar
				class = "DRUID",
				rankLevels = {10, 20, 32, 42, 52, 62, 71, 77},
				amounts = {31, 51, 68, 103, 138, 220, 299, 408},
				spec = "Feral",
				improvedPercentagePerTalentPoint = 8,
				maxTalentPoints = 5,
				improved = GetSpellInfo(16862)				-- Feral Aggression
			},
			[GetSpellInfo(47437)] = {						-- Demoralizing Shout
				class = "WARRIOR",
				rankLevels = {14, 24, 34, 44, 54, 62, 70, 79},
				amounts = {35, 55, 70, 105, 140, 220, 300, 418},
				spec = "Fury",
				improvedPercentagePerTalentPoint = 8,
				maxTalentPoints = 5,
				improved = GetSpellInfo(12879)				-- Improved Demoralizing Shout
			},
			[GetSpellInfo(55487)] = {						-- Demoralizing Screech
				class = "HUNTER",
				rankLevels = {1, 16, 32, 48, 64, 80},
				amounts = {25, 45, 70, 100, 210, 410},
				pet = L["Bird of Prey"],
				improved = GetSpellInfo(12879)
			},
		},
	},
	[L["Bleed Damage"]] = {
		icon = select(3, GetSpellInfo(48564)),
		spells = {
			[GetSpellInfo(48564)] = {						-- Mangle - Bear
				class = "DRUID",
				amount = 30,
				amountType = "%",
				spec = "Feral",
				requiredTalent = GetSpellInfo(33917),		-- Mangle
			},
			[GetSpellInfo(48566)] = {						-- Mangle - Cat
				class = "DRUID",
				amount = 30,
				amountType = "%",
				spec = "Feral",
				requiredTalent = GetSpellInfo(33917),		-- Mangle
			},
			[GetSpellInfo(46855)] = {						-- Trauma
				class = "WARRIOR",
				amount = 30,
				amountType = "%",
				spec = "Arms",
				requiredTalent = GetSpellInfo(46855),		-- Trauma
			},
		},
	},
	[L["Cast Speed Slow"]] = {
		icon = select(3, GetSpellInfo(5761)),
		spells = {
			[GetSpellInfo(11719)] = {						-- Curse of Tongues
				class = "WARLOCK",
				rankLevels = {26, 50},
				amounts = {25, 30},
				amountType = "%"
			},
			[GetSpellInfo(5761)] = {						-- Mind-numbing Poison
				class = "ROGUE",
				amount = 30,
				amountType = "%"
			},
			[GetSpellInfo(58611)] = {						-- Lava Breath
				class = "HUNTER",
				amount = 25,
				amountType = "%",
				spec = "Beastmaster",
				requiredTalent = GetSpellInfo(53270),		-- Beast Mastery
				pet = L["Core Hound"]
			},
			[GetSpellInfo(31589)] = {						-- Slow
				class = "MAGE",
				spec = "Arcane",
				requiredTalent = GetSpellInfo(31589),		-- Slow
				amount = 30,
				amountType = "%"
			},
		},
	},
	[L["Critical Strike Chance Taken"]] = {
		icon = select(3, GetSpellInfo(58410)),
		spells = {
			[GetSpellInfo(20337)] = {						-- Heart of the Crusader
				class = "PALADIN",
				spec = "Retribution",
				maxTalentPoints = 3,
				amountPerTalentPoint = 1,
				amountType = "%",
				requiredTalent = GetSpellInfo(20337)		-- Heart of the Crusader
			},
			[GetSpellInfo(58410)] = {						-- Master Poisoner
				class = "ROGUE",
				spec = "Assassination",
				maxTalentPoints = 3,
				amountPerTalentPoint = 1,
				amountType = "%",
				requiredTalent = GetSpellInfo(58410)		-- Master Poisoner
			},
			[GetSpellInfo(30706)] = {						-- Totem of Wrath
				class = "SHAMAN",
				totem = "fire",
				spec = "Elemental",
				amount = 3,
				amountType = "%",
				requiredTalent = GetSpellInfo(30706)		-- Totem of Wrath
			},
		},
	},
	[L["Healing Taken"]] = {
		icon = select(3, GetSpellInfo(47486)),
		spells = {
			[GetSpellInfo(49050)] = {						-- Aimed Shot
				class = "HUNTER",
				spec = "Marksmanship",
				amount = 50,
				amountType = "%",
				requiredTalent = GetSpellInfo(49050)		-- Aimed Shot
			},
			[GetSpellInfo(46911)] = {						-- Furious Attacks
				class = "WARRIOR",
				spec = "Fury",
				amountPerStack = 25,
				maxStacks = 2,
				amountType = "%",
				requiredTalent = GetSpellInfo(46911)		-- Furious Attacks
			},
			[GetSpellInfo(47486)] = {						-- Mortal Strike
				class = "WARRIOR",
				spec = "Arms",
				amount = 50,
				amountType = "%",
				requiredTalent = GetSpellInfo(47486)		-- Mortal Strike
			},
			[GetSpellInfo(13218)] = {						-- Wound Poison (unranked name)
				class = "ROGUE",
				amount = 50,
				amountType = "%"
			},
		},
	},
	[L["Health Restore"]] = {
		icon = select(3, GetSpellInfo(20271)),
		spells = {
			[GetSpellInfo(20271)] = {						-- Judgement of Light
				class = "PALADIN",
				exclusiveWith = L["Mana Restore"],
			},
		},
	},
	[L["Mana Restore"]] = {
		icon = select(3, GetSpellInfo(53408)),
		spells = {
			[GetSpellInfo(53408)] = {						-- Judgement of Wisdom
				class = "PALADIN",
				exclusiveWith = L["Health Restore"]
			},
		},		
	},
	[L["Melee Attack Speed Slow"]] = {
		icon = select(3, GetSpellInfo(47502)),
		spells = {
			[GetSpellInfo(55095)] = {						-- Frost Fever
				class = "DEATHKNIGHT",
				spec = "Frost",
				amount = 14,
				amountType = "%",
				source = GetSpellInfo(49909),				-- Icy Touch
				improvedAmountPerTalentPoint = 2,
				maxTalentPoints = 3,
				improved = GetSpellInfo(51456)				-- Improved Icy Touch
			},
			[GetSpellInfo(48485)] = {						-- Infected Wounds
				class = "DRUID",
				spec = "Feral",
				maxStacks = 2,
				amountPerTalentPoint = 10/3,
				amountType = "%",
				maxTalentPoints = 3,
				requiredTalent = GetSpellInfo(48485)		-- Infected Wounds
			},
			[GetSpellInfo(53696)] = {						-- Judgements of the Just
				class = "PALADIN",
				spec = "Protection",
				amountPerTalentPoint = 10,
				amountType = "%",
				maxTalentPoints = 2,
				requiredTalent = GetSpellInfo(53696)		-- Judgements of the Just
			},
			[GetSpellInfo(47502)] = {						-- Thunder Clap
				class = "WARRIOR",
				spec = "Protection",
				amount = 10,
				amountType = "%",
				maxTalentPoints = 3,
				improvedAmountPerTalentPoint = 10/3,
				improved = GetSpellInfo(12666)				-- Improved Thunder Clap
			},
		},
	},
	[L["Melee Hit Chance Reduction"]] = {
		icon = select(3, GetSpellInfo(48468)),
		spells = {
			[GetSpellInfo(48468)] = {						-- Insect Swarm
				class = "DRUID",
				amount = 3,
				amountType = "%",
				requiredTalent = GetSpellInfo(48468),
			},
			[GetSpellInfo(3043)] = {						-- Scorpid Sting
				class = "HUNTER",
				amount = 3,
				amountType = "%"
			},
		},
	},
	[L["Physical Vulnerability"]] = {
		icon = select(3, GetSpellInfo(58413)),
		spells = {
			[GetSpellInfo(29859)] = {						-- Blood Frenzy
				class = "WARRIOR",
				spec = "Arms",
				amountPerTalentPoint = isWoW3dot1 and 2 or 1,
				amountType = "%",
				maxTalentPoints = 2,
				requiredTalent = GetSpellInfo(29859)		-- Blood Frenzy
			},
			[GetSpellInfo(58413)] = {						-- Savage Combat
				class = "ROGUE",
				spec = "Combat",
				amountPerTalentPoint = isWoW3dot1 and 2 or 1,
				amountType = "%",
				maxTalentPoints = 2,
				requiredTalent = GetSpellInfo(58413)		-- Savage Combat
			},
		},
	},
	[L["Spell Critical Strike Chance"]] = {
		icon = select(3, GetSpellInfo(12873)),
		spells = {
			[GetSpellInfo(12873)] = {						-- Improved Scorch
				class = "MAGE",
				spec = "Fire",
				amountPerStack = isWoW3dot1 and 1 or 2,
				maxStacks = 5,
				amountType = "%",
				maxTalentPoints = 5,
				requiredTalent = GetSpellInfo(12873)
			},
			[GetSpellInfo(28593)] = {						-- Winter's Chill
				class = "MAGE",
				spec = "Frost",
				amountPerStack = isWoW3dot1 and 1 or 2,
				maxStacks = 5,
				amountType = "%",
				maxTalentPoints = 5,
				requiredTalent = GetSpellInfo(28593)
			},
			[GetSpellInfo(17800)] = {
				class = "WARLOCK",
				spec = "Destruction",
				amount= 5,
				amountType = "%",
				maxTalentPoints = 5,
				requiredTalent = GetSpellInfo(17803)
			},
		},
	},
	[L["Spell Damage Taken"]] = {
		icon = select(3, GetSpellInfo(47865)),
		spells = {
			[GetSpellInfo(47865)] = {						-- Curse of the Elements
				class = "WARLOCK",
				spec = "Affliction",
				amount = 13,
				amountType = "%"
			},
			[GetSpellInfo(48511)] = {						-- Earth and Moon
				class = "DRUID",
				spec = "Balance",
				amountPerTalentPoint = 13/3,
				amountType = "%",
				maxTalentPoints = 3,
				requiredTalent = GetSpellInfo(48511)		-- Earth and Moon
			},
			[GetSpellInfo(51735)] = {						-- Ebon Plague
				class = "DEATHKNIGHT",
				spec = "Unholy",
				amountPerTalentPoint = 13/3,
				amountType = "%",
				maxTalentPoints = 3,
				requiredTalent = GetSpellInfo(51161)		-- Ebon Plaguebringer
			},
		},
	},
	[L["Spell Hit Chance Taken"]] = {
		icon = select(3, GetSpellInfo(33193)),
		spells = {
			[GetSpellInfo(33602)] = {						-- Improved Faerie Fire
				class = "DRUID",
				spec = "Balance",
				amountPerTalentPoint = 1,
				amountType = "%",
				maxTalentPoints = 3,
				requiredTalent = GetSpellInfo(33602)		-- Improved Faerie Fire
			},
			[GetSpellInfo(33193)] = {						-- Misery
				class = "PRIEST",
				spec = "Shadow",
				amount = 3,
				amountType = "%",
				requiredTalent = GetSpellInfo(33193)		-- Misery
			},
		},
	},
}

-- Mapping spells. These don't actually appear on the target, but are
-- 'hidden' auras that we have to infer, often given as an additional
-- effect to a spell.  (With the exception of Wound Poison which is a
-- straight mapping to cope with the varying names)

Utopia.mapping = {
	[GetSpellInfo(13218)] = {					-- Wound Poison (unranked name)
		class = "ROGUE",
		type = "map",
		list = {
			[GetSpellInfo(13219)] = true,		-- Wound Poison I
			[GetSpellInfo(13225)] = true,		-- Wound Poison II
			[GetSpellInfo(13226)] = true,		-- Wound Poison III
			[GetSpellInfo(13227)] = true,		-- Wound Poison IV
			[GetSpellInfo(27188)] = true,		-- Wound Poison V
			[GetSpellInfo(57977)] = true,		-- Wound Poison VI
			[GetSpellInfo(57978)] = true,		-- Wound Poison VII
		},
	},

	[GetSpellInfo(58410)] = {					-- Master Poisoner
		class = "ROGUE",
		requiredTalent = GetSpellInfo(58410),	-- Master Poisoner
		type = "trigger",
		list = {
			[GetSpellInfo(3408)] = true,		-- Crippling Poison
			[GetSpellInfo(41190)] = true,		-- Mind-numbing Poison
			[GetSpellInfo(13218)] = true,		-- Wound Poison
			[GetSpellInfo(2823)] = true,		-- Deadly Poison
		},
	},

	[GetSpellInfo(53696)] = {					-- Judgements of the Just
		class = "PALADIN",
		requiredTalent = GetSpellInfo(53696),	-- Judgements of the Just
		type = "trigger",
		list = {
			[GetSpellInfo(53408)] = true,		-- Judgement of Wisdom
			[GetSpellInfo(20271)] = true,		-- Judgement of Light
			[GetSpellInfo(53407)] = true,		-- Judgement of Justice
		},
	},

	[GetSpellInfo(20140)] = {					-- Improved Devotion Aura
		class = "PALADIN",
		requiredTalent = GetSpellInfo(20140),	-- Improved Devotion Aura
		type = "trigger",
		list = {
			[GetSpellInfo(48942)] = true,		-- Devotion Aura
		},
	},

	[GetSpellInfo(33182)] = {					-- Improved Divine Spirit
		class = "PRIEST",
		requiredTalent = GetSpellInfo(33182),	-- Improved Divine Spirit
		type = "trigger",
		list = {
			[GetSpellInfo(48073)] = true,		-- Divine Spirit
		},
	},

	[GetSpellInfo(53648)] = {					-- Swift Retribution
		class = "PALADIN",
		requiredTalent = GetSpellInfo(53648),	-- Swift Retribution
		type = "trigger",
		list = {
			[GetSpellInfo(48942)] = true,		-- Devotion Aura
			[GetSpellInfo(54043)] = true,		-- Retribution Aura
			[GetSpellInfo(19746)] = true,		-- Concentration Aura
			[GetSpellInfo(48943)] = true,		-- Shadow Resistance Aura
			[GetSpellInfo(48945)] = true,		-- Frost Resistance Aura
			[GetSpellInfo(48947)] = true,		-- Fire Resistance Aura
			[GetSpellInfo(32223)] = true,		-- Crusader Aura
		},
	},

	[GetSpellInfo(31869)] = {					-- Sanctified Retribution
		class = "PALADIN",
		requiredTalent = GetSpellInfo(31869),	-- Sanctified Retribution
		type = "trigger",
		list = {
			[GetSpellInfo(48942)] = true,		-- Devotion Aura
			[GetSpellInfo(54043)] = true,		-- Retribution Aura
			[GetSpellInfo(19746)] = true,		-- Concentration Aura
			[GetSpellInfo(48943)] = true,		-- Shadow Resistance Aura
			[GetSpellInfo(48945)] = true,		-- Frost Resistance Aura
			[GetSpellInfo(48947)] = true,		-- Fire Resistance Aura
			[GetSpellInfo(32223)] = true,		-- Crusader Aura
		},
	},
}

for Type,data in pairs(Utopia.debuffs) do
	assert(type(Type) == "string")
	assert(data.spells)
	for spellName,info in pairs(data.spells) do
		assert(info.class)
		assert(type(spellName) == "string")
		if (info.amount) then
			assert(type(info.amount) == "number")
		elseif (info.amountsPerStack) then
			assert(info.rankLevels)
			assert(#info.amountsPerStack == #info.rankLevels)
		elseif (info.amounts) then
			assert(type(info.amounts) == "table")
			assert(info.rankLevels)
			assert(#info.amounts == #info.rankLevels)
		end
	end
end

for name,data in pairs(Utopia.mapping) do
	assert(data.class)
	assert(data.list)
	for spellName,True in pairs(data.list) do
		assert(type(spellName) == "string")
		assert(True)
	end
end	
